﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace lab2
{
    enum CubeType
    {
        Singular,
        D,
        D_fault_1,
        D_fault_0
    }

    enum Symbol
    {
        X,
        D,
        notD,
        One,
        Zero,
        Empty
    }

    ///////////////////////////////////////////////////////////////////////////

    class Element
    {
        protected List<Symbol[]> cubesList = new List<Symbol[]>();
        private int currentCubeIndex;
        private Symbol[] savedCube = new Symbol[13];

        public Element()
        {
            currentCubeIndex = 0;
        }

        // Получить следующий куб элемента
        public bool getNextCube(ref Symbol[] cube)
        {
            if (currentCubeIndex < cubesList.Count)
            {
                cube = cubesList[currentCubeIndex];
                ++currentCubeIndex;
                return true;
            }
            else            
                currentCubeIndex = 0;
            
            return false;
        }

        // Установить куб элемента в кубе схемы
        public void setFullCubes(List<Symbol[]> cubesBuffer, int cubeIndex)
        {
            int i = 0;
            int count = cubesBuffer.Count;

            initializeEmptyCubes(count);

            switch (cubeIndex)
            {
                case 1:
                    for (i = 0; i < count; i++)
                    {
                        cubesList[i][0] = cubesBuffer[i][0];
                        cubesList[i][1] = cubesBuffer[i][1];
                        cubesList[i][7] = cubesBuffer[i][2];
                    }
                    break;
                case 2:
                    for (i = 0; i < count; i++)
                    {
                        cubesList[i][2] = cubesBuffer[i][0];
                        cubesList[i][8] = cubesBuffer[i][1];
                    }
                    break;
                case 3:
                    for (i = 0; i < count; i++)
                    {
                        cubesList[i][4] = cubesBuffer[i][0];
                        cubesList[i][5] = cubesBuffer[i][1];
                        cubesList[i][9] = cubesBuffer[i][2];
                    }
                    break;
                case 4:
                    for (i = 0; i < count; i++)
                    {
                        cubesList[i][3] = cubesBuffer[i][0];
                        cubesList[i][9] = cubesBuffer[i][1];
                        cubesList[i][6] = cubesBuffer[i][2];
                        cubesList[i][10] = cubesBuffer[i][3];
                    }
                    break;
                case 5:
                    for (i = 0; i < count; i++)
                    {
                        cubesList[i][8] = cubesBuffer[i][0];
                        cubesList[i][10] = cubesBuffer[i][1];
                        cubesList[i][11] = cubesBuffer[i][2];
                    }
                    break;
                case 6:
                    for (i = 0; i < count; i++)
                    {
                        cubesList[i][7] = cubesBuffer[i][0];
                        cubesList[i][11] = cubesBuffer[i][1];
                        cubesList[i][12] = cubesBuffer[i][2];
                    }
                    break;
            }
        }

        public Symbol[] getSavedCube()
        {
            return savedCube;
        }

        public void setSavedCube(Symbol[] cube)
        {
            savedCube = cube;
        }

        private void initializeEmptyCubes(int count)
        {
            for (int i = 0; i < count; i++)
            {
                Symbol[] cube = new Symbol[13];
                for (int j = 0; j < cube.Length; j++)                
                    cube[j] = Symbol.X;

                cubesList.Add(cube);
            }
        }

        // Все кубы данного элемента были перебраны в ходе алгоритма
        public Boolean isLastCube()
        {
            return currentCubeIndex == cubesList.Count;
        }

        // Пересечение символов сингулярных кубов
        public static Symbol IntersectionOfSingularSymbols(Symbol Symbol1, Symbol Symbol2)
        {
            if (Symbol1 == Symbol2 || Symbol2 == Symbol.X)            
                return Symbol1;
            
            if (Symbol1 == Symbol.X)            
                return Symbol2;            

            if (Symbol1 == Symbol.Zero && Symbol2 == Symbol.One)            
                return Symbol.notD;
            
            if (Symbol1 == Symbol.One && Symbol2 == Symbol.Zero)            
                return Symbol.D;
            
            return Symbol.X;
        }

        public static Symbol[] GetDCub(Symbol[] n1, Symbol[] n2)
        {
            if (n1.Length != n2.Length)
                return null;

            Symbol[] resultCube = new Symbol[n1.Length];
            for (int i = 0; i < n1.Length; i++)            
                resultCube[i] = IntersectionOfSingularSymbols(n1[i], n2[i]);
            
            return resultCube;
        }

    }

    ///////////////////////////////////////////////////////////////////////////

    class F1 : Element
    {
        public F1(CubeType type)
        {
            List<Symbol[]> reducedCubesList = new List<Symbol[]>();

            switch (type)
            {
                case CubeType.Singular:
                    reducedCubesList.Add(new Symbol[3] { Symbol.Zero, Symbol.One, Symbol.Zero });
                    reducedCubesList.Add(new Symbol[3] { Symbol.One, Symbol.Zero, Symbol.Zero });
                    reducedCubesList.Add(new Symbol[3] { Symbol.Zero, Symbol.Zero, Symbol.One });
                    reducedCubesList.Add(new Symbol[3] { Symbol.One, Symbol.One, Symbol.One });
                    break;
                case CubeType.D:
                    List<Symbol[]> tmp_cubesList = new List<Symbol[]>();
                    tmp_cubesList.Add(new Symbol[3] { Symbol.Zero, Symbol.One, Symbol.Zero });
                    tmp_cubesList.Add(new Symbol[3] { Symbol.One, Symbol.Zero, Symbol.Zero });
                    tmp_cubesList.Add(new Symbol[3] { Symbol.Zero, Symbol.Zero, Symbol.One });
                    tmp_cubesList.Add(new Symbol[3] { Symbol.One, Symbol.One, Symbol.One });
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[0], tmp_cubesList[2]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[0], tmp_cubesList[3]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[1], tmp_cubesList[2]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[1], tmp_cubesList[3]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[2], tmp_cubesList[0]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[2], tmp_cubesList[1]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[3], tmp_cubesList[0]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[3], tmp_cubesList[1]));
                    break;
                case CubeType.D_fault_1:
                    reducedCubesList.Add(new Symbol[3] { Symbol.Zero, Symbol.One, Symbol.notD });
                    reducedCubesList.Add(new Symbol[3] { Symbol.One, Symbol.Zero, Symbol.notD });
                    break;
                case CubeType.D_fault_0:
                    reducedCubesList.Add(new Symbol[3] { Symbol.Zero, Symbol.Zero, Symbol.D });
                    reducedCubesList.Add(new Symbol[3] { Symbol.One, Symbol.One, Symbol.D });
                    break;
            }

            setFullCubes(reducedCubesList, 1);
        }
    }

    ///////////////////////////////////////////////////////////////////////////

    class F2 : Element
    {
        public F2(CubeType type)
        {
            List<Symbol[]> reducedCubesList = new List<Symbol[]>();

            switch (type)
            {
                case CubeType.Singular:
                    reducedCubesList.Add(new Symbol[2] { Symbol.Zero, Symbol.One });
                    reducedCubesList.Add(new Symbol[2] { Symbol.One, Symbol.Zero });
                    break;
                case CubeType.D:
                    List<Symbol[]> tmp_cubesList = new List<Symbol[]>();
                    tmp_cubesList.Add(new Symbol[2] { Symbol.Zero, Symbol.One });
                    tmp_cubesList.Add(new Symbol[2] { Symbol.One, Symbol.Zero });
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[0], tmp_cubesList[1]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[1], tmp_cubesList[0]));
                    break;
                case CubeType.D_fault_1:
                    reducedCubesList.Add(new Symbol[2] { Symbol.One, Symbol.notD });
                    break;
                case CubeType.D_fault_0:
                    reducedCubesList.Add(new Symbol[2] { Symbol.Zero, Symbol.D });
                    break;
            }

            setFullCubes(reducedCubesList, 2);
        }
    }

    ///////////////////////////////////////////////////////////////////////////

    class F3 : Element
    {
        public F3(CubeType type)
        {
            List<Symbol[]> reducedCubesList = new List<Symbol[]>();

            switch (type)
            {
                case CubeType.Singular:
                    reducedCubesList.Add(new Symbol[3] { Symbol.One, Symbol.One, Symbol.Zero });
                    reducedCubesList.Add(new Symbol[3] { Symbol.X, Symbol.Zero, Symbol.One });
                    reducedCubesList.Add(new Symbol[3] { Symbol.Zero, Symbol.X, Symbol.One });
                    break;
                case CubeType.D:
                    List<Symbol[]> tmp_cubesList = new List<Symbol[]>();
                    tmp_cubesList.Add(new Symbol[3] { Symbol.One, Symbol.One, Symbol.Zero });
                    tmp_cubesList.Add(new Symbol[3] { Symbol.X, Symbol.Zero, Symbol.One });
                    tmp_cubesList.Add(new Symbol[3] { Symbol.Zero, Symbol.X, Symbol.One });
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[0], tmp_cubesList[1]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[0], tmp_cubesList[2]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[1], tmp_cubesList[0]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[2], tmp_cubesList[0]));
                    break;
                case CubeType.D_fault_1:
                    reducedCubesList.Add(new Symbol[3] { Symbol.One, Symbol.One, Symbol.notD });
                    break;
                case CubeType.D_fault_0:
                    reducedCubesList.Add(new Symbol[3] { Symbol.X, Symbol.Zero, Symbol.D });
                    reducedCubesList.Add(new Symbol[3] { Symbol.Zero, Symbol.X, Symbol.D });
                    break;
            }

            setFullCubes(reducedCubesList, 3);
        }
    }

    ///////////////////////////////////////////////////////////////////////////

    class F4 : Element
    {
        public F4(CubeType type)
        {
            List<Symbol[]> reducedCubesList = new List<Symbol[]>();

            switch (type)
            {
                case CubeType.Singular:
                    reducedCubesList.Add(new Symbol[4] { Symbol.Zero, Symbol.Zero, Symbol.Zero, Symbol.One });
                    reducedCubesList.Add(new Symbol[4] { Symbol.X, Symbol.One, Symbol.X, Symbol.Zero });
                    reducedCubesList.Add(new Symbol[4] { Symbol.X, Symbol.X, Symbol.One, Symbol.Zero });
                    reducedCubesList.Add(new Symbol[4] { Symbol.One, Symbol.X, Symbol.X, Symbol.Zero });
                    break;
                case CubeType.D:
                    List<Symbol[]> tmp_cubesList = new List<Symbol[]>();
                    tmp_cubesList.Add(new Symbol[4] { Symbol.Zero, Symbol.Zero, Symbol.Zero, Symbol.One });
                    tmp_cubesList.Add(new Symbol[4] { Symbol.X, Symbol.One, Symbol.X, Symbol.Zero });
                    tmp_cubesList.Add(new Symbol[4] { Symbol.X, Symbol.X, Symbol.One, Symbol.Zero });
                    tmp_cubesList.Add(new Symbol[4] { Symbol.One, Symbol.X, Symbol.X, Symbol.Zero });
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[3], tmp_cubesList[0]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[2], tmp_cubesList[0]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[1], tmp_cubesList[0]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[0], tmp_cubesList[3]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[0], tmp_cubesList[2]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[0], tmp_cubesList[1]));
                    break;
                case CubeType.D_fault_1:
                    reducedCubesList.Add(new Symbol[4] { Symbol.One, Symbol.X, Symbol.X, Symbol.notD });
                    reducedCubesList.Add(new Symbol[4] { Symbol.X, Symbol.One, Symbol.X, Symbol.notD });
                    reducedCubesList.Add(new Symbol[4] { Symbol.X, Symbol.X, Symbol.One, Symbol.notD });
                    break;
                case CubeType.D_fault_0:
                    reducedCubesList.Add(new Symbol[4] { Symbol.Zero, Symbol.Zero, Symbol.Zero, Symbol.D });
                    break;
            }

            setFullCubes(reducedCubesList, 4);

        }
    }

    ///////////////////////////////////////////////////////////////////////////

    class F5 : Element
    {
        public F5(CubeType type)
        {
            List<Symbol[]> reducedCubesList = new List<Symbol[]>();

            switch (type)
            {
                case CubeType.Singular:
                    reducedCubesList.Add(new Symbol[3] { Symbol.Zero, Symbol.Zero, Symbol.Zero });
                    reducedCubesList.Add(new Symbol[3] { Symbol.X, Symbol.One, Symbol.One });
                    reducedCubesList.Add(new Symbol[3] { Symbol.One, Symbol.X, Symbol.One });
                    break;
                case CubeType.D:
                    List<Symbol[]> tmp_cubesList = new List<Symbol[]>();
                    tmp_cubesList.Add(new Symbol[3] { Symbol.Zero, Symbol.Zero, Symbol.Zero });
                    tmp_cubesList.Add(new Symbol[3] { Symbol.X, Symbol.One, Symbol.One });
                    tmp_cubesList.Add(new Symbol[3] { Symbol.One, Symbol.X, Symbol.One });
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[0], tmp_cubesList[1]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[0], tmp_cubesList[2]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[1], tmp_cubesList[0]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[2], tmp_cubesList[0]));
                    break;
                case CubeType.D_fault_1:
                    reducedCubesList.Add(new Symbol[3] { Symbol.Zero, Symbol.Zero, Symbol.notD });
                    break;
                case CubeType.D_fault_0:
                    reducedCubesList.Add(new Symbol[3] { Symbol.X, Symbol.One, Symbol.D });
                    reducedCubesList.Add(new Symbol[3] { Symbol.One, Symbol.X, Symbol.D });
                    break;
            }

            setFullCubes(reducedCubesList, 5);
        }
    }

    ///////////////////////////////////////////////////////////////////////////

    class F6 : Element
    {
        public F6(CubeType type)
        {
            List<Symbol[]> reducedCubesList = new List<Symbol[]>();

            switch (type)
            {
                case CubeType.Singular:
                    reducedCubesList.Add(new Symbol[3] { Symbol.Zero, Symbol.X, Symbol.Zero });
                    reducedCubesList.Add(new Symbol[3] { Symbol.X, Symbol.Zero, Symbol.Zero });
                    reducedCubesList.Add(new Symbol[3] { Symbol.One, Symbol.One, Symbol.One });
                    break;
                case CubeType.D:
                    List<Symbol[]> tmp_cubesList = new List<Symbol[]>();
                    tmp_cubesList.Add(new Symbol[3] { Symbol.Zero, Symbol.X, Symbol.Zero });
                    tmp_cubesList.Add(new Symbol[3] { Symbol.X, Symbol.Zero, Symbol.Zero });
                    tmp_cubesList.Add(new Symbol[3] { Symbol.One, Symbol.One, Symbol.One });
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[0], tmp_cubesList[2]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[1], tmp_cubesList[2]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[2], tmp_cubesList[0]));
                    reducedCubesList.Add(Element.GetDCub(tmp_cubesList[2], tmp_cubesList[1]));
                    break;
                case CubeType.D_fault_1:
                    reducedCubesList.Add(new Symbol[3] { Symbol.Zero, Symbol.X, Symbol.notD });
                    reducedCubesList.Add(new Symbol[3] { Symbol.X, Symbol.Zero, Symbol.notD });
                    break;
                case CubeType.D_fault_0:
                    reducedCubesList.Add(new Symbol[3] { Symbol.One, Symbol.One, Symbol.D });
                    break;
            }

            setFullCubes(reducedCubesList, 6);
        }
    }
}
