﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace lab3
{
    struct Cover
    {
        public bool[] inputs;       // Входной набор
        public bool[] faultCover;   // Покрытие неисправностей
    }


    class Tests
    {
        private int inputsNumber;
        private int nodesCount;
        private int faultCount;

        //------------------------------------------------------------------

        public Cover[] coversArray { get; private set; }
        public Cover[] minCoversArray { get; private set; }

        public bool[][] defaultNodesStatesArray { get; private set; }
        public bool[][] optimalNodesStatesArray { get; private set; }

        public int switchCountBad { get; private set; }
        public int switchCountOptimal { get; private set; }
        public int minTestNumber { get; private set; }

        //-------------------------------------------------------------------

        public Tests(int inputsNumber, int nodesCount)
        {
            this.inputsNumber = inputsNumber;
            this.nodesCount = nodesCount;
            this.faultCount = nodesCount * 2;
        }


        public void FindAll()
        {
            coversArray = GetCoverForEveryTest();         // Получаем покрытия для всех входных комбинаций

            minCoversArray = GetMinimumFullTestsSet(coversArray);           // Получаем мин. число тестов для полного покрытия

            defaultNodesStatesArray = new bool[minCoversArray.Length][];
            for (int i = 0; i < minCoversArray.Length; i++)
                defaultNodesStatesArray[i] = GetAllNodesState(minCoversArray[i].inputs);

            // Получаем оптимальную перестановку тестов для минимального числа переключений
            optimalNodesStatesArray = GetOptimalPermutationOfNodesStatesArray(defaultNodesStatesArray);

            switchCountBad = GetFullSwitchesCount(defaultNodesStatesArray);
            switchCountOptimal = GetFullSwitchesCount(optimalNodesStatesArray);
            minTestNumber = minCoversArray.Length;
        }


        private bool[][] GetOptimalPermutationOfNodesStatesArray(bool[][] nodesStatesArray)
        {
            int[] indexes = Enumerable.Range(0, nodesStatesArray.Length).ToArray();

            int[] optimalPermutation = (int[])indexes.Clone();
            int minSwitchCount = GetFullSwitchesCount(nodesStatesArray, indexes);

            while (Permutations.NextPermutation(indexes) == true)
            {
                int switchCount = GetFullSwitchesCount(nodesStatesArray, indexes);
                if (switchCount < minSwitchCount)
                {
                    minSwitchCount = switchCount;
                    optimalPermutation = (int[])indexes.Clone();
                }
            }

            bool[][] result = new bool[nodesStatesArray.Length][];
            for (int i = 0; i < result.Length; i++)
                result[i] = nodesStatesArray[optimalPermutation[i]];

            return result;
        }


        private Cover[] GetCoverForEveryTest()
        {
            int combinationsNumber = (int)Math.Pow(2, inputsNumber);

            Cover[] coversArray = new Cover[combinationsNumber];

            for (int i = 0; i < combinationsNumber; i++)
            {
                coversArray[i].inputs = GetBoolArrayFromInt(i, inputsNumber);
                coversArray[i].faultCover = GetFaultCover(coversArray[i].inputs, faultCount);
            }
            return coversArray;
        }


        private Cover[] GetMinimumFullTestsSet(Cover[] coversArray)
        {
            List<Cover> minCoversList = new List<Cover>(50);
            bool[] currentCover = new bool[faultCount];

            while (isFullCover(currentCover) == false)       // Пока не покрыты все нисправности
            {
                int maxIndex = 0;
                int maxCount = 0;
                for (int i = 0; i < coversArray.Length; i++)
                {
                    int coveredFaultsCount = 0;             // Число покрытых неисправностей данным тестовым набором
                    for (int j = 0; j < nodesCount; j++)
                    {
                        // Если данный набор покрывает данную неисправность const 0 и она ещё не покрыта
                        if (currentCover[j] == false && coversArray[i].faultCover[j] == true)
                            coveredFaultsCount++;
                        // Если данный набор покрывает данную неисправность const 1 и она ещё не покрыта
                        if (currentCover[j + nodesCount] == false && coversArray[i].faultCover[j + nodesCount] == true)
                            coveredFaultsCount++;
                    }
                    if (coveredFaultsCount > maxCount)
                    {
                        maxCount = coveredFaultsCount;
                        maxIndex = i;
                    }
                }

                GetFaultCover(currentCover, coversArray[maxIndex].faultCover);  // Обновить currentCover

                Cover cover = new Cover { inputs = coversArray[maxIndex].inputs, faultCover = (bool[])currentCover.Clone() };
                minCoversList.Add(cover);
            }

            return minCoversList.ToArray();
        }


        // Получить состояние всех полюсов схемы
        private bool[] GetAllNodesState(bool[] inputs)
        {
            bool[] nodesState = new bool[nodesCount];

            for (int i = 0; i < inputs.Length; i++)
                nodesState[i] = inputs[i];

            nodesState[7] = !(inputs[0] ^ inputs[1]);                   // F1
            nodesState[8] = !inputs[2];                                 // F2
            nodesState[9] = !(inputs[4] & inputs[5]);                   // F3
            nodesState[10] = !(inputs[3] | nodesState[9] | inputs[6]);  // F4
            nodesState[11] = nodesState[8] | nodesState[10];            // F5
            nodesState[12] = nodesState[7] & nodesState[11];            // F6

            return nodesState;
        }


        private int GetFullSwitchesCount(bool[][] nodesStatesArray, int[] indexes)
        {
            bool[] prevStates = nodesStatesArray[indexes[0]];
            bool[] newStates;

            int fullSwitchCount = 0;
            for (int i = 1; i <= nodesStatesArray.Length; i++)
            {
                if (i == nodesStatesArray.Length)
                    newStates = nodesStatesArray[indexes[0]];
                else
                    newStates = nodesStatesArray[indexes[i]];

                fullSwitchCount += GetSwitchesCount(prevStates, newStates);

                prevStates = newStates;
            }
            return fullSwitchCount;
        }


        static private int GetFullSwitchesCount(bool[][] nodesStatesArray)
        {
            bool[] prevStates = nodesStatesArray[0];
            bool[] newStates;

            int fullSwitchCount = 0;
            for (int i = 1; i <= nodesStatesArray.Length; i++)
            {
                if (i == nodesStatesArray.Length)
                    newStates = nodesStatesArray[0];
                else
                    newStates = nodesStatesArray[i];

                fullSwitchCount += GetSwitchesCount(prevStates, newStates);

                prevStates = newStates;
            }
            return fullSwitchCount;
        }


        static public int GetSwitchesCount(bool[] states1, bool[] states2)
        {
            int count = 0;
            for (int i = 0; i < states1.Length; i++)
            {
                if (states1[i] != states2[i])
                    count++;
            }
            return count;
        }



        //--------------------------------------------------------
        //--------------------------------------------------------

        static private bool[] GetBoolArrayFromInt(int num, int returnBitsCount = 32)
        {
            if (returnBitsCount > sizeof(int) * 8)
                returnBitsCount = sizeof(int) * 8;

            bool[] bits = new bool[returnBitsCount];
            for (int i = 0; i < returnBitsCount; i++)
                bits[i] = (num & (1 << i)) != 0;

            return bits;
        }



        static private bool isFullCover(bool[] cover)
        {
            for (var i = 0; i < cover.Length; i++)
            {
                if (cover[i] == false)
                    return false;
            }
            return true;
        }


        static private void GetFaultCover(bool[] currentCover, bool[] testCover)
        {
            if (currentCover.Length != testCover.Length)
                throw new Exception("Covers must have same length!");

            for (int i = 0; i < currentCover.Length; i++)
                currentCover[i] = currentCover[i] || testCover[i];
        }


        static private bool[] GetFaultCover(bool[] inputs, int faultsNum)
        {
            bool correctResult = CalculateSet(inputs);

            bool[] cover = new bool[faultsNum];
            int nodesCount = cover.Length / 2;

            for (int i = 0; i < nodesCount; i++)
            {
                if (CalculateSet(inputs, i, false) != correctResult)
                    cover[i] = true;

                if (CalculateSet(inputs, i, true) != correctResult)
                    cover[nodesCount + i] = true;
            }
            return cover;
        }


        // Получить значение на выходе схемы
        static private bool CalculateSet(bool[] inputs, int faultNumber = -1, bool faultType = false)
        {
            bool tmp = false;
            if (faultNumber >= 0 && faultNumber <= 6)
            {
                tmp = inputs[faultNumber];
                inputs[faultNumber] = faultType;
            }

            bool f1 = !(inputs[0] ^ inputs[1]);
            bool f2 = !inputs[2];
            bool f3 = !(inputs[4] & inputs[5]);


            bool f4, f5, f6;

            if (faultNumber == 9) f4 = !(inputs[3] | faultType | inputs[6]);
            else f4 = !(inputs[3] | f3 | inputs[6]);


            if (faultNumber == 8) f5 = faultType | f4;
            else if (faultNumber == 10) f5 = f2 | faultType;
            else f5 = f2 | f4;

            if (faultNumber == 11) f6 = f1 & faultType;
            else if (faultNumber == 7) f6 = faultType & f5;
            else f6 = f1 & f5;

            if (faultNumber == 12)
                f6 = faultType;

            if (faultNumber >= 0 && faultNumber <= 6)   // Возвращаем изменённое значение бита (если меняли)
                inputs[faultNumber] = tmp;

            return f6;
        }
    }
}
