﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace lab5_LFSR
{
    struct State
    {
        public bool[] CoverState;
        public bool[] CurrentState;
    }


    public partial class Form1 : Form
    {
        private const int InputsNumber = 7;
        private const int FaultCount = 26;

        private readonly int[] _polynomial = { 9, 5 };  // Здесь задаётся полином 
        private readonly int _lfsrLength;

        public Form1()
        {
            _lfsrLength = (_polynomial[0] > InputsNumber) ? _polynomial[0] : InputsNumber;

            InitializeComponent();

            StringBuilder sb = new StringBuilder(30);
            foreach (int t in _polynomial)
            {
                sb.Append("x^");
                sb.Append(t.ToString());
                sb.Append('⊕');
            }
            sb.Append('1');
            label_polynomial.Text = sb.ToString();
        }

        private void button_start_Click(object sender, EventArgs e)
        {
            int combinationsNumber = (int)Math.Pow(2, _lfsrLength);
            listView1.Items.Clear();

            List<List<State>> tests = new List<List<State>>();

            int minCount = int.MaxValue;

            // Перебор всех начальных состояний LFSR
            for (int i = 0; i < combinationsNumber; i++)
            {
                bool[] initLfsrState = new bool[_lfsrLength];
                new BitArray(new int[] { i }) { Length = _lfsrLength }.CopyTo(initLfsrState, 0);

                // Получение списка тактов LFSR для полного покрытия
                List<State> statesList = GetLfsrStatesList(initLfsrState, minCount);

                if (statesList == null)
                    continue;

                if (statesList.Count < minCount)
                {
                    tests.Clear();
                    tests.Add(statesList);
                    minCount = statesList.Count;
                }
                else if (statesList.Count == minCount)
                    tests.Add(statesList);                                 
            }

            textBox_min.Text = minCount.ToString();
            textBox_minTestNum.Text = tests.Count.ToString();

            foreach (var coversList in tests)
            {
                for (int i = 0; i < coversList.Count; i++)
                {
                    string input = BoolArrayToString(coversList[i].CurrentState);
                    string cover = BoolArrayToString(coversList[i].CoverState);

                    listView1.Items.Add(new ListViewItem(new[] {(i+1).ToString(), input, cover}));
                }
                listView1.Items.Add(new ListViewItem(new[] { "", "", "" }));
            }
        }


        private List<State> GetLfsrStatesList(bool[] initLfsrState, int maxLength = int.MaxValue)
        {
            if (Array.TrueForAll(initLfsrState, y => y == false))
                return null;

            List<State> coversList = new List<State>();
            bool[] cover = new bool[FaultCount];

            LFSR lfsr = new LFSR(_polynomial, initLfsrState);

            while (IsFullCover(cover) == false)
            {
                bool[] currentInput = lfsr.GetSubState(InputsNumber);
                cover = GetFaultCover(currentInput, cover);

                State state = new State
                {
                    CoverState = cover,
                    CurrentState = (bool[])lfsr.State.Clone()
                    //CurrentState = currentInput
                };

                coversList.Add(state);

                if (coversList.Count > maxLength)
                    return null;

                lfsr.NextState();
            }
            return coversList;
        }
        

        private static bool IsFullCover(IEnumerable<bool> cover)
        {
            return cover.All(t => t != false);
        }


        private static bool[] GetFaultCover(IList<bool> inputs, bool[] currentCover)
        {
            bool correctResult = CalculateSet(inputs);

            bool[] cover = (bool[])currentCover.Clone();
            int nodesCount = cover.Length / 2;

            for (int i = 0; i < nodesCount; i++)
            {
                if (CalculateSet(inputs, i, false) != correctResult)
                    cover[i] = true;

                if (CalculateSet(inputs, i, true) != correctResult)
                    cover[nodesCount + i] = true;
            }
            return cover;
        }

        private static bool CalculateSet(IList<bool> inputs, int faultNumber = -1, bool faultType = false)
        {
            bool tmp = false;
            if (faultNumber >= 0 && faultNumber <= 6)
            {
                tmp = inputs[faultNumber];
                inputs[faultNumber] = faultType; 
            }
                
            bool f1 = !(inputs[0] ^ inputs[1]);
            bool f2 = !inputs[2];
            bool f3 = !(inputs[4] & inputs[5]);


            bool f4, f5, f6;

            if (faultNumber == 9)      f4 = !(inputs[3] | faultType | inputs[6]);
            else                        f4 = !(inputs[3] | f3 | inputs[6]);


            if (faultNumber == 8)       f5 = faultType | f4;
            else if (faultNumber == 10) f5 = f2 | faultType;
            else                        f5 = f2 | f4;

            if (faultNumber == 11)      f6 = f1 & faultType;
            else if (faultNumber == 7)  f6 = faultType & f5;
            else                        f6 = f1 & f5;

            if (faultNumber == 12)
                f6 = faultType;

            if (faultNumber >= 0 && faultNumber <= 6)   // Возвращаем изменённое значение бита (если меняли)
                inputs[faultNumber] = tmp;

            return f6;
        }


        private static string BoolArrayToString(IEnumerable<bool> array)
        {
            var sb = new StringBuilder();
            foreach (bool t in array)
            {
                sb.Append(t ? "1 " : "0 ");
            }

            return sb.ToString();
        }
    }
}
