﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace lab5_LFSR
{
    struct State
    {
        public bool[] cover_state;
        public bool[] LFSR_state;
    }


    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();

            lfsr = new LFSR(new int[] { 6, 4, 2, 0 });
        }

        private LFSR lfsr;

        private const int InputsNumber = 7;
        private const int faultCount = 26;


        private void button_start_Click(object sender, EventArgs e)
        {
            int combinationsNumber = (int) Math.Pow(2, InputsNumber);
            listView1.Items.Clear();

            List<List<State>> tests = new List<List<State>>();

            int minCount = int.MaxValue;

            // Перебор всех начальных состояний LFSR
            for (byte i = 0; i < combinationsNumber; i++)
            {
                bool[] inputsArray = new bool[InputsNumber];
                new BitArray(new byte[] {i}) {Length = InputsNumber} .CopyTo(inputsArray, 0);

                // Получение списка тактов LFSR для полного покрытия
                List<State> statesList = GetLFSRStatesList(inputsArray, minCount);

                if (statesList == null)
                    continue;

                if (statesList.Count < minCount)
                {
                    tests.Clear();
                    tests.Add(statesList);
                    minCount = statesList.Count;
                }
                else if (statesList.Count == minCount)
                    tests.Add(statesList);                                 
            }

            textBox_min.Text = minCount.ToString();
            textBox_minTestNum.Text = tests.Count.ToString();

            foreach (var coversList in tests)
            {
                for (int i = 0; i < coversList.Count; i++)
                {
                    string input = BoolArrayToString(coversList[i].LFSR_state);
                    string cover = BoolArrayToString(coversList[i].cover_state);

                    listView1.Items.Add(new ListViewItem(new[] {(i+1).ToString(), input, cover}));
                }
                listView1.Items.Add(new ListViewItem(new[] { "", "", "" }));
            }
        }


        private List<State> GetLFSRStatesList(bool[] inputsArray, int maxLength = int.MaxValue)
        {
            List<State> coversList = new List<State>();
            bool[] cover = new bool[faultCount];

            while (isFullCover(cover) == false)
            {
                cover = GetFaultCover(inputsArray, cover);

                State state = new State
                {
                    cover_state = cover,
                    LFSR_state = inputsArray
                };

                coversList.Add(state);

                if (coversList.Count > maxLength)
                    return null;

                bool[] nextState = lfsr.GetNext(inputsArray);
                if (nextState.SequenceEqual(inputsArray))
                    return null;

                inputsArray = nextState;
            }
            return coversList;
        }
        

        private bool isFullCover(bool[] cover)
        {
            for (var i = 0; i < cover.Length; i++)
            {
                if (cover[i] == false)                
                    return false;                
            }
            return true;
        }


        private bool[] GetFaultCover(bool[] inputs, bool[] currentCover)
        {
            bool correctResult = CalculateSet(inputs);

            bool[] cover = (bool[])currentCover.Clone();
            int nodesCount = cover.Length / 2;

            for (int i = 0; i < nodesCount; i++)
            {
                if (CalculateSet(inputs, i, false) != correctResult)
                    cover[i] = true;

                if (CalculateSet(inputs, i, true) != correctResult)
                    cover[nodesCount + i] = true;
            }
            return cover;
        }

        private bool CalculateSet(bool[] inputs, int faultNumber = -1, bool faultType = false)
        {
            bool tmp = false;
            if (faultNumber >= 0 && faultNumber <= 6)
            {
                tmp = inputs[faultNumber];
                inputs[faultNumber] = faultType; 
            }
                
            bool f1 = !(inputs[0] ^ inputs[1]);
            bool f2 = !inputs[2];
            bool f3 = !(inputs[4] & inputs[5]);


            bool f4, f5, f6;

            if (faultNumber == 9)      f4 = !(inputs[3] | faultType | inputs[6]);
            else                        f4 = !(inputs[3] | f3 | inputs[6]);


            if (faultNumber == 8)       f5 = faultType | f4;
            else if (faultNumber == 10) f5 = f2 | faultType;
            else                        f5 = f2 | f4;

            if (faultNumber == 11)      f6 = f1 & faultType;
            else if (faultNumber == 7)  f6 = faultType & f5;
            else                        f6 = f1 & f5;

            if (faultNumber == 12)
                f6 = faultType;

            if (faultNumber >= 0 && faultNumber <= 6)   // Возвращаем изменённое значение бита (если меняли)
                inputs[faultNumber] = tmp;

            return f6;
        }


        private string BoolArrayToString(bool[] array)
        {
            var sb = new StringBuilder();
            for (var i = 0; i < array.Length; i++)
            {
                sb.Append(array[i] ? "1 " : "0 ");
            }

            return sb.ToString();
        }
    }
}
