﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace kidsvt_lab5_SA
{
    class LFSR
    {
        private int[] _polynomial_XOR_members;
        private bool[] _initialState;

        public bool[] State { get; private set; }


        public LFSR(int[] polynomial_XOR_members, bool[] initialState)
        {
            _polynomial_XOR_members = new int[polynomial_XOR_members.Length];
            for (int i = 0; i < _polynomial_XOR_members.Length; i++)
            {
                _polynomial_XOR_members[i] = polynomial_XOR_members[i] - 1;
                if (_polynomial_XOR_members[i] < 0 || _polynomial_XOR_members[i] >= initialState.Length)
                    throw new Exception("Wrong polynomial or initial state length!");
            }

            State = initialState;
            _initialState = (bool[])initialState.Clone();
        }

        public void Restart()
        {
            State = (bool[])_initialState.Clone();
        }

        public void NextState()
        {
            bool firstValue = CalculateFirstValue(State);

            for (int i = State.Length - 1; i >= 1; i--)            
                State[i] = State[i - 1];

            State[0] = firstValue;
        }


        private bool CalculateFirstValue(bool[] curState)
        {
            return _polynomial_XOR_members.Aggregate(false, (current, member) => current ^ curState[member]);
        }
    }
}
