﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace kidsvt_lab5_SA
{
    struct SignatureAnalyzerPolynomialInfo
    {
        public int[] polynomial;
        public double percent;
        public int XOR_count;
    }

    
    class SignatureAnalyzer
    {
        const int nodesNumber = 13;                 // Количество полюсов схемы
        const int faultNumber = nodesNumber * 2;    // Количество неисправностей


        private int[] _polynomial_XOR_members;

        public bool[] State { get; private set; }


        public SignatureAnalyzer(int polynomialSize, int[] polynomial_XOR_members)
        {
            _polynomial_XOR_members = new int[polynomial_XOR_members.Length];
            for (int i = 0; i < _polynomial_XOR_members.Length; i++)
            {
                _polynomial_XOR_members[i] = polynomial_XOR_members[i] - 1;
                if (_polynomial_XOR_members[i] < 0 || _polynomial_XOR_members[i] >= polynomialSize)
                    throw new Exception("Incorrect polynomial!");
            }

            State = Enumerable.Repeat(false, polynomialSize).ToArray();
        }

        private void Restart()
        {
            State = Enumerable.Repeat(false, State.Length).ToArray();
        }

        private void NextState(bool deviceOutput)
        {
            bool firstValue = CalculateFirstValue(State, deviceOutput);

            for (int i = State.Length - 1; i >= 1; i--)            
                State[i] = State[i - 1];

            State[0] = firstValue;
        }


        private bool CalculateFirstValue(bool[] curState, bool deviceOutput)
        {
            bool tmp = _polynomial_XOR_members.Aggregate(false, (current, member) => current ^ curState[member]);
            return tmp ^ deviceOutput;
        }


        private static bool[] GetSignatureAnalyzerValue(SignatureAnalyzer sa, LFSR lfsr, int cyclesCount)
        {
            lfsr.Restart();
            sa.Restart();

            // Get Signature Analyzer Reference Value
            for (int i = 0; i < cyclesCount; i++)
            {
                bool deviceOutput = Device.GetDeviceOutput(lfsr.State);
                sa.NextState(deviceOutput);
                lfsr.NextState();
            }
            bool[] referenceSignatureAnalyzerValue = (bool[])sa.State.Clone();
            return referenceSignatureAnalyzerValue;
        }

        private static bool[] GetSignatureAnalyzerValue(SignatureAnalyzer sa, LFSR lfsr, int cyclesCount, 
                                                                         int faultNodeNumber, bool faultType)
        {
            lfsr.Restart();
            sa.Restart();

            // Get Signature Analyzer Value
            for (int i = 0; i < cyclesCount; i++)
            {
                bool deviceOutput = Device.GetDeviceOutput(lfsr.State, faultNodeNumber, faultType);
                sa.NextState(deviceOutput);
                lfsr.NextState();
            }
            bool[] referenceSignatureAnalyzerValue = (bool[])sa.State.Clone();
            return referenceSignatureAnalyzerValue;
        }
        
        //-----------------------------------------------------------
        //-----------------------------------------------------------


        public static LinkedList<SignatureAnalyzerPolynomialInfo> GetSignatureAnalyzerStatistics(List<int[]> PolynomList, LFSR lfsr,
                                                    int signatureAnalyzer_length, int cyclesCount)
        {
            LinkedList<SignatureAnalyzerPolynomialInfo> infoList = new LinkedList<SignatureAnalyzerPolynomialInfo>();
            foreach (int[] currentPolynomial in PolynomList)
            {
                int coverFaultsCount = 0;

                SignatureAnalyzer signatureAnalyzer =
                    new SignatureAnalyzer(signatureAnalyzer_length, currentPolynomial);

                // Get Signature Analyzer Reference Value
                bool[] referenceSignatureAnalyzerValue =
                    SignatureAnalyzer.GetSignatureAnalyzerValue(signatureAnalyzer, lfsr, cyclesCount);

                // Compare Signature Analyzer Value on faults with Reference Value
                for (int i = 0; i < nodesNumber; i++)
                {
                    // const 0 fault
                    bool[] signatureAnalyzerValue =
                        SignatureAnalyzer.GetSignatureAnalyzerValue(signatureAnalyzer, lfsr, cyclesCount, i, false);
                    if (signatureAnalyzerValue.SequenceEqual(referenceSignatureAnalyzerValue) == false)
                        coverFaultsCount++;

                    // const 1 fault
                    signatureAnalyzerValue =
                        SignatureAnalyzer.GetSignatureAnalyzerValue(signatureAnalyzer, lfsr, cyclesCount, i, true);
                    if (signatureAnalyzerValue.SequenceEqual(referenceSignatureAnalyzerValue) == false)
                        coverFaultsCount++;
                }
                double coverPercent = (double)coverFaultsCount / (double)faultNumber * 100.0;

                SignatureAnalyzerPolynomialInfo info = new SignatureAnalyzerPolynomialInfo
                {
                    polynomial = currentPolynomial,
                    percent = (double)coverFaultsCount / (double)faultNumber * 100.0,
                    XOR_count = currentPolynomial.Length
                };
                infoList.AddLast(info);
            }
            return infoList;
        }
    }
}
