﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace kidsvt_lab6_SA
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
            button_generateSequence_Click(this, null);
        }

        const int sequenceLength = 40;

        const int signatureAnalyzer_length = 8;
        int[] signatureAnalyzer_polynomial = { 8, 6, 5, 1 };

        const int startErrorMultiplicity = 1;
        const int endErrorMultiplicity = 4;


        struct SA_ErrorCoverInfo
        {
            public int errorMultiplicity;      // Кратность ошибки

            public int errorsCount;            // Возможное количество ошибок с данной кратностью
            public int coveredErrorsCount;     // Количество покрытых ошибок с данной кратностью

            public bool[][] uncoveredErrors;   // Непокрытые ошибки
        }

        private void ClearControls()
        {
            dataGridView1.Rows.Clear();
            dataGridView1.Update();
            treeView_uncoveredErrors.Nodes.Clear();
        }


        private void start_button_Click(object sender, EventArgs e)
        {
            //var watch = Stopwatch.StartNew();
            try
            {
                start_button.Enabled = false;

                ClearControls();

                bool[] inputSequence = StringToBool(textBox_sequence.Text);

                SignatureAnalyzer sa = new SignatureAnalyzer(signatureAnalyzer_length,
                                                                    signatureAnalyzer_polynomial);
                SignatureAnalyzer_2channel sa2 = new SignatureAnalyzer_2channel(2, signatureAnalyzer_length);

                var statistics1 = Get_SA_ErrorsCoverStatistics(startErrorMultiplicity, endErrorMultiplicity, sa, inputSequence);
                var statistics2 = Get_SA_ErrorsCoverStatistics_2channel(startErrorMultiplicity, endErrorMultiplicity, sa2, inputSequence);

                ShowStatistics("Одноканальный СА", statistics1);
                ShowStatistics("Двухканальный СА", statistics2);                
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, ex.GetType().ToString(), MessageBoxButtons.OK, MessageBoxIcon.Warning);
            }
            finally
            {
                start_button.Enabled = true;
            }
            //watch.Stop();
            //MessageBox.Show(watch.ElapsedMilliseconds.ToString());
        }
            



        private void ShowStatistics(string statName, LinkedList<SA_ErrorCoverInfo> statistics)
        {
            dataGridView1.Rows.Add(new[] { "", statName, "", "" });
            TreeNode treeNode_SAchannel = new TreeNode(statName);

            int coveredErrorsCount = 0;
            int errorsCount = 0;
            foreach (var info in statistics)
            {
                coveredErrorsCount += info.coveredErrorsCount;
                errorsCount += info.errorsCount;

                string errType = info.errorMultiplicity.ToString() + "-кратные ошибки";
                double percent = (double)info.coveredErrorsCount / (double)info.errorsCount * 100.0;

                dataGridView1.Rows.Add(new object[] { errType, info.errorsCount, info.coveredErrorsCount, percent });                

                if (info.uncoveredErrors.Length > 0)
                {
                    TreeNode treeNode_errorType = new TreeNode(errType);
                    treeNode_SAchannel.Nodes.Add(treeNode_errorType);

                    for (int i = 0; i < info.uncoveredErrors.Length; i++)
                        treeNode_errorType.Nodes.Add(BoolArrayToString(info.uncoveredErrors[i]));
                }                               
            }
            double generalPercent = (double)coveredErrorsCount / (double)errorsCount * 100.0;
            dataGridView1.Rows.Add(new object[] { "Всего", errorsCount, coveredErrorsCount, generalPercent });

            if (coveredErrorsCount < errorsCount)
                treeView_uncoveredErrors.Nodes.Add(treeNode_SAchannel);
        }
        


        private LinkedList<SA_ErrorCoverInfo> Get_SA_ErrorsCoverStatistics(int startErrorMult, int endErrorMult, 
                                                                                            SignatureAnalyzer sa, bool[] inputSequence)
        {
            Object lockMe = new Object();

            int[] sequenceIndexes = Enumerable.Range(0, inputSequence.Length).ToArray();

            bool[][] correctSignatures = SignatureAnalyzer.GetAllSignatures(sa, inputSequence);

            LinkedList<SA_ErrorCoverInfo> statistics = new LinkedList<SA_ErrorCoverInfo>();

            for (int i = startErrorMult; i <= endErrorMult; i++)
            {
                int errorsCount = 0;
                List<bool[]> uncoveredErrors = new List<bool[]>(500);

                var combinations = Combinations.GetCombinations<int>(sequenceIndexes, i);

                Parallel.ForEach(combinations, indexesCombination =>
                {
                    SignatureAnalyzer s = new SignatureAnalyzer(sa);

                    bool[] errorSequence = (bool[])inputSequence.Clone();
                    for (int j = 0; j < indexesCombination.Length; j++)
                        errorSequence[indexesCombination[j]] = !errorSequence[indexesCombination[j]];

                    // Сигнатуры совпадают - необнаруженная ошибка
                    int lastInd = indexesCombination[indexesCombination.Length - 1];
                    if (SignatureAnalyzer.CompareSignatures(s, errorSequence, correctSignatures[lastInd], lastInd))
                    {
                        lock (lockMe)
                        {
                            uncoveredErrors.Add(errorSequence);
                        }                        
                    }
                    Interlocked.Increment(ref errorsCount);

                });

                SA_ErrorCoverInfo coverInfo = new SA_ErrorCoverInfo
                {
                    errorMultiplicity = i,
                    errorsCount = errorsCount,
                    coveredErrorsCount = errorsCount - uncoveredErrors.Count,
                    uncoveredErrors = uncoveredErrors.ToArray()
                };

                statistics.AddLast(coverInfo);
            }
            return statistics;
        }

        private LinkedList<SA_ErrorCoverInfo> Get_SA_ErrorsCoverStatistics_2channel(int startErrorMult, int endErrorMult,
                                                                                            SignatureAnalyzer_2channel sa, bool[] inputSequence)
        {
            Object lockMe = new Object();
            int[] sequenceIndexes = Enumerable.Range(0, inputSequence.Length).ToArray();

            bool[][] correctSignatures = SignatureAnalyzer_2channel.GetAllSignatures(sa, inputSequence);

            LinkedList<SA_ErrorCoverInfo> statistics = new LinkedList<SA_ErrorCoverInfo>();

            for(int i = startErrorMult; i <= endErrorMult; i++)
            {
                int errorsCount = 0;
                List<bool[]> uncoveredErrors = new List<bool[]>(500);

                var combinations = Combinations.GetCombinations<int>(sequenceIndexes, i);

                Parallel.ForEach(combinations, indexesCombination =>
                {
                    SignatureAnalyzer_2channel s = new SignatureAnalyzer_2channel(sa);

                    bool[] errorSequence = (bool[])inputSequence.Clone();
                    for (int j = 0; j < indexesCombination.Length; j++)
                        errorSequence[indexesCombination[j]] = !errorSequence[indexesCombination[j]];

                    // Сигнатуры совпадают - необнаруженная ошибка
                    int lastInd = indexesCombination[indexesCombination.Length - 1];
                    if (SignatureAnalyzer_2channel.CompareSignatures(s, errorSequence, correctSignatures[lastInd / sa.channelNumber], lastInd))
                    {
                        lock (lockMe)
                        {
                            uncoveredErrors.Add(errorSequence);
                        }
                    }

                    Interlocked.Increment(ref errorsCount);
                });

                SA_ErrorCoverInfo coverInfo = new SA_ErrorCoverInfo
                {
                    errorMultiplicity = i,
                    errorsCount = errorsCount,
                    coveredErrorsCount = errorsCount - uncoveredErrors.Count,
                    uncoveredErrors = uncoveredErrors.ToArray()
                };

                statistics.AddLast(coverInfo);
            }
            return statistics;
        }

        private string BoolArrayToString(bool[] boolArray)
        {
            StringBuilder sb = new StringBuilder(boolArray.Length);
            for (int i = 0; i < boolArray.Length; i++)
            {
                if (boolArray[i] == true)
                    sb.Append('1');
                else
                    sb.Append('0');
            }
            return sb.ToString();
        }

        private bool[] StringToBool(string str)
        {
            bool[] result = new bool[str.Length];

            for (int i = 0; i < str.Length; i++)
            {
                if (str[i] == '1')
                    result[i] = true;
                else if (str[i] == '0')
                    result[i] = false;
                else
                    throw new Exception("Bad input string format! Only '1' and '0' allowed!");
            }
            return result;
        }

        private void textBox_sequence_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (!(e.KeyChar == '1' || e.KeyChar == '0' || char.IsControl(e.KeyChar)))
            {
                e.Handled = true;
            }
        }

        private void button_generateSequence_Click(object sender, EventArgs e)
        {
            Random random = new Random();
            StringBuilder sb = new StringBuilder(sequenceLength);

            for (int i = 0; i < sequenceLength; i++)
            {
                if (random.NextDouble() < 0.5)
                    sb.Append('0');
                else
                    sb.Append('1');
            }

            textBox_sequence.Text = sb.ToString();
        }
    
    }
}
