﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace kidsvt_lab6_SA
{    
    class SignatureAnalyzer
    {
        
        private int[] _polynomial_XOR_members;

        public bool[] State { get; private set; }


        public SignatureAnalyzer(int polynomialSize, int[] polynomial_XOR_members)
        {
            _polynomial_XOR_members = new int[polynomial_XOR_members.Length];
            for (int i = 0; i < _polynomial_XOR_members.Length; i++)
            {
                _polynomial_XOR_members[i] = polynomial_XOR_members[i] - 1;
                if (_polynomial_XOR_members[i] < 0 || _polynomial_XOR_members[i] >= polynomialSize)
                    throw new Exception("Incorrect polynomial!");
            }

            State = Enumerable.Repeat(false, polynomialSize).ToArray();
        }

        public SignatureAnalyzer(SignatureAnalyzer sa)
        {
            _polynomial_XOR_members = (int[])sa._polynomial_XOR_members.Clone();
            State = (bool[])sa.State.Clone();
        }

        private void Restart()
        {
            State = Enumerable.Repeat(false, State.Length).ToArray();
        }

        private void NextState(bool deviceOutput)
        {
            bool firstValue = CalculateFirstValue(State, deviceOutput);

            for (int i = State.Length - 1; i >= 1; i--)            
                State[i] = State[i - 1];

            State[0] = firstValue;
        }


        private bool CalculateFirstValue(bool[] curState, bool deviceOutput)
        {
            bool tmp = _polynomial_XOR_members.Aggregate(false, (current, member) => current ^ curState[member]);
            return tmp ^ deviceOutput;
        }
        
        //-----------------------------------------------------------
        //-----------------------------------------------------------

        public static bool[][] GetAllSignatures(SignatureAnalyzer sa, bool[] inputSequence)
        {
            bool[][] signatures = new bool[inputSequence.Length][];

            sa.Restart();

            for (int i = 0; i < inputSequence.Length; i++)
            {
                sa.NextState(inputSequence[i]);
                signatures[i] = (bool[])sa.State.Clone();
            }

            return signatures;
        }

        public static bool[] GetSignature(SignatureAnalyzer sa, bool[] inputSequence)
        {
            sa.Restart();

            for (int i = 0; i < inputSequence.Length; i++)            
                sa.NextState(inputSequence[i]);

            return (bool[])sa.State.Clone();
        }

        public static bool[] GetSignature(SignatureAnalyzer sa, bool[] inputSequence, int index)
        {
            sa.Restart();

            if (index < 0 || index >= inputSequence.Length)
                throw new Exception("Bad index for signature!");

            for (int i = 0; i < index; i++)
                sa.NextState(inputSequence[i]);

            return (bool[])sa.State.Clone();
        }


        public static bool CompareSignatures(SignatureAnalyzer sa, bool[] inputSequence1, 
                                                                            bool[] inputSequence2)
        {
            sa.Restart();
            for (int i = 0; i < inputSequence1.Length; i++)
                sa.NextState(inputSequence1[i]);

            bool[] sign1 = (bool[])sa.State.Clone();

            sa.Restart();
            for (int i = 0; i < inputSequence2.Length; i++)
                sa.NextState(inputSequence2[i]);

            bool[] sign2 = (bool[])sa.State.Clone();

            return sign1.SequenceEqual(sign2);
        }


        public static bool CompareSignatures(SignatureAnalyzer sa, bool[] inputSequence,
                                                                      bool[] signature, int index)
        {
            if (index < 0 || index >= inputSequence.Length)
                throw new Exception("Bad index for signature!");

            sa.Restart();
            for (int i = 0; i <= index; i++)
                sa.NextState(inputSequence[i]);

            bool[] signature2 = (bool[])sa.State.Clone();

            return signature2.SequenceEqual(signature);
        }
    }
}
