﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace kidsvt_lab6_SA
{
    class SignatureAnalyzer_2channel
    {
        private int[] _polynomial_XOR_members;

        public bool[] State { get; private set; }

        public int channelNumber { get; private set; }

        public SignatureAnalyzer_2channel(int channelNumber, int polynomialSize = 8, int[] polynomial_XOR_members = null)
        {
            State = Enumerable.Repeat(false, polynomialSize).ToArray();
            this.channelNumber = channelNumber;
        }

        public SignatureAnalyzer_2channel(SignatureAnalyzer_2channel sa)
        {
            //_polynomial_XOR_members = (int[])sa._polynomial_XOR_members.Clone();
            State = (bool[])sa.State.Clone();
            channelNumber = sa.channelNumber;
        }

        private void Restart()
        {
            State = Enumerable.Repeat(false, State.Length).ToArray();
        }

        private void NextState(bool y, bool y_1)
        {
            bool firstValue = State[0] ^ State[3] ^ State[5] ^ State[6] ^ State[7] ^ y_1;
            bool secondValue = State[0] ^ State[4] ^ State[5] ^ State[7] ^ y;

            for (int i = State.Length - 1; i >= 2; i--)            
                State[i] = State[i - 2];

            State[0] = firstValue;
            State[1] = secondValue;
        }

        //------------------------------------------------------------------------------------
        //------------------------------------------------------------------------------------
        //------------------------------------------------------------------------------------


        public static bool[][] GetAllSignatures(SignatureAnalyzer_2channel sa, bool[] inputSequence)
        {
            int siganturesCount = SequencePadding(ref inputSequence, sa.channelNumber);
            bool[][] signatures = new bool[siganturesCount][];

            sa.Restart();

            for (int i = 0; i < inputSequence.Length; i += sa.channelNumber)
            {
                sa.NextState(inputSequence[i], inputSequence[i + 1]);
                signatures[i / sa.channelNumber] = (bool[])sa.State.Clone();
            }

            return signatures;
        }

        public static bool[] GetSignature(SignatureAnalyzer_2channel sa, bool[] inputSequence)
        {
            sa.Restart();
            SequencePadding(ref inputSequence, sa.channelNumber);

            for (int i = 0; i < inputSequence.Length; i += sa.channelNumber)            
                sa.NextState(inputSequence[i], inputSequence[i + 1]);            

            return (bool[])sa.State.Clone();
        }


        public static bool[] GetSignature(SignatureAnalyzer_2channel sa, bool[] inputSequence, int index)
        {
            sa.Restart();
            SequencePadding(ref inputSequence, sa.channelNumber);

            if (index < 0 || index >= inputSequence.Length)
                throw new Exception("Bad index for signature!");

            for (int i = 0; i < index; i += sa.channelNumber)
                sa.NextState(inputSequence[i], inputSequence[i + 1]);

            return (bool[])sa.State.Clone();
        }


        public static bool CompareSignatures(SignatureAnalyzer_2channel sa, bool[] inputSequence1,
                                                                                bool[] inputSequence2)
        {
            SequencePadding(ref inputSequence1, sa.channelNumber);
            SequencePadding(ref inputSequence2, sa.channelNumber);

            sa.Restart();
            for (int i = 0; i < inputSequence1.Length; i += sa.channelNumber)
                sa.NextState(inputSequence1[i], inputSequence1[i + 1]);

            bool[] sign1 = (bool[])sa.State.Clone();

            sa.Restart();
            for (int i = 0; i < inputSequence2.Length; i += sa.channelNumber)
                sa.NextState(inputSequence2[i], inputSequence2[i + 1]);

            bool[] sign2 = (bool[])sa.State.Clone();

            return sign1.SequenceEqual(sign2);
        }


        public static bool CompareSignatures(SignatureAnalyzer_2channel sa, bool[] inputSequence,
                                                                      bool[] signature, int index)
        {
            SequencePadding(ref inputSequence, sa.channelNumber);

            if (index < 0 || index >= inputSequence.Length)
                throw new Exception("Bad index for signature!");

            sa.Restart();
            for (int i = 0; i <= index; i += sa.channelNumber)
                sa.NextState(inputSequence[i], inputSequence[i + 1]);

            bool[] signature2 = (bool[])sa.State.Clone();

            return signature2.SequenceEqual(signature);
        }

        private static int SequencePadding(ref bool[] inputSequence, int channelCount)
        {
            int mod = inputSequence.Length % channelCount;
            int siganturesCount = inputSequence.Length / channelCount;
            if (mod != 0)
            {
                siganturesCount++;
                Array.Resize(ref inputSequence, siganturesCount * channelCount);
            }
            return siganturesCount;
        }
    }
}
