﻿using System;
using System.Collections.Generic;
using System.Threading.Tasks;
using System.Windows.Forms;
using kidsvt_lab7_8.Enums;

namespace kidsvt_lab7_8
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        private readonly Func<MemoryUnit, int>[] _testMethods1 = { MemoryTests.Galpat, MemoryTests.Mats };
        private readonly FaultType[] _faultTypes1 = {FaultType.AF, FaultType.SAF, FaultType.CFin, FaultType.CFid};

        private readonly Func<MemoryUnit, int>[] _testMethods2 = { MemoryTests.Mats, MemoryTests.March_PS };
        private readonly FaultType[] _faultTypes2 = { FaultType.PNPSFK3, FaultType.ANPSFK5 };

        private const int NumberOfMemoryRows = 64;
        private const int NumberOfMemoryColumns = 64;
        private const int FaultsCount = 512;

        private void ClearControls()
        {
            dataGridView1.Rows.Clear();
            dataGridView1.Update();            
        }


        private void start_button_Click(object sender, EventArgs e)
        {
            //var watch = Stopwatch.StartNew();
            try
            {
                start_button.Enabled = false;

                ClearControls();

                GetTestsStatistics(_testMethods1, _faultTypes1);
                GetTestsStatistics(_testMethods2, _faultTypes2);

            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, ex.GetType().ToString(), MessageBoxButtons.OK, MessageBoxIcon.Warning);
            }
            finally
            {
                start_button.Enabled = true;
            }
            //watch.Stop();
            //MessageBox.Show(watch.ElapsedMilliseconds.ToString());
        }


        private void GetTestsStatistics(IEnumerable<Func<MemoryUnit, int>> testMethods, FaultType[] faultTypes)
        {
            foreach (var testMethod in testMethods)
            {
                Func<MemoryUnit, int> method = testMethod;

                Parallel.ForEach(faultTypes, faultType =>
                {
                    var memoryUnit = new MemoryUnit(NumberOfMemoryRows, NumberOfMemoryColumns,
                        faultType, FaultsCount);

                    int foundFaultsCount = method(memoryUnit);
                    string testName = GetTestFuncName(method).ToUpper();

                    int faultsCount = (faultType == FaultType.AF)
                        ? memoryUnit.AddressFaultsCount
                        : memoryUnit.FaultsCount;

                    double coverPercent = foundFaultsCount / (double)faultsCount * 100.0;

                    object[] row = { testName, faultType.ToString(), coverPercent };

                    BeginInvoke(new Action(() =>
                    {
                        dataGridView1.Rows.Add(row);
                        dataGridView1.Update();
                    }));  
                });

                BeginInvoke(new Action(() =>
                {
                    dataGridView1.Rows.Add(new object[] { "", "", "" });
                    dataGridView1.Update();
                })); 
            }
        }


        private string GetTestFuncName(Func<MemoryUnit, int> func)
        {
            return func.Method.Name;
        }
    }
}
